import os, tempfile, json, pickle
import param
from lancet.core import PrettyPrinted

try:    import numpy
except: pass

try:    import Image
except: pass

class FileType(PrettyPrinted, param.Parameterized):
    """
    The base class for all supported file types in Lancet. This class
    is designed to be simple and easily extensible to support new
    files and has only three essential methods: 'save', 'data' and
    'metadata').
    """

    hash_suffix = param.Boolean(default=True, doc='''
       Whether to ensure the saved filename is unique by adding a
       short hash suffix. Note that this is a class level parameter
       only.''')

    directory = param.String(default='.', allow_None=True, doc='''
       Directory in which to load or save the file. Note that this
       is a class level parameter only.''')

    extensions = param.List(default=[], constant=True,
       doc= '''The set of supported file extensions.''')

    data_key = param.String(default='data', doc='''
       The name (key) given to the file contents if the key cannot be
       determined from the file itself.''')

    def __init__(self, **params):
        super(FileType, self).__init__(**params)
        self._pprint_args = ([],[],None,{})
        self.pprint_args(['data_key', 'hash_suffix'], [])

    def save(self, filename,  metadata={}, **data):
        """
        The implementation in the base class simply checks there is no
        clash between the metadata and data keys.
        """
        intersection = set(metadata.keys()) & set(data.keys())
        if intersection:
            msg = 'Key(s) overlap between data and metadata: %s'
            raise Exception(msg  % ','.join(intersection))

    def metadata(self, filename):
        """
        The metadata returned as a dictionary.
        """
        raise NotImplementedError

    def data(self, filename):
        """
        Data returned as a dictionary.
        """
        raise NotImplementedError

    def _loadpath(self, filename):
        (_, ext) = os.path.splitext(filename)
        if ext not in self.extensions:
            raise Exception("Unsupported extensions")
        abspath = os.path.abspath(os.path.join(self.directory, filename))
        return filename if os.path.isfile(filename) else abspath

    def _savepath(self, filename):
        """
        Returns the full path for saving the file, adding an extension
        and making the filename unique as necessary.
        """
        (basename, ext) = os.path.splitext(filename)
        basename = basename if (ext in self.extensions) else filename
        ext = ext if (ext in self.extensions) else self.extensions[0]
        savepath = os.path.abspath(os.path.join(self.directory,
                                                 '%s%s' % (basename, ext)))
        return (tempfile.mkstemp(ext, basename + "_", self.directory)[1]
                if self.hash_suffix else savepath)

    @classmethod
    def file_supported(cls, filename):
        """
        Returns a boolean indicating whether the filename has an
        appropriate extension for this class.
        """
        if not isinstance(filename, str):
            return False
        (_, ext) = os.path.splitext(filename)
        if ext not in cls.extensions:
            return False
        else:
            return True

    def __repr__(self):
        return self._pprint(flat=True, annotate=False)

    def __str__(self):
        return self._pprint(annotate=False)

    def __or__(self, other):
        return FileOption(self, other)



class FileOption(FileType):
    """
    Allows a FileType out of a pair of FileTypes to handle a given
    file. For instance, given a mixed list of .png image filenames and
    .npz Numpy filenames, the ImageFile() | NumpyFile() object an
    handle either type of filename appropriately.
    """
    first = param.ClassSelector(class_=FileType, doc='''
       The first potential FileType to handle a given filename.''')

    second = param.ClassSelector(class_=FileType, doc='''
       The second potential FileType to handle a given filename.''')

    def __init__(self, first, second, **params):
        if set(first.extensions) & set(second.extensions):
            raise Exception("FileTypes must support non-overlapping sets of extensions.")
        extensions = set(first.extensions) | set(second.extensions)
        super(FileOption, self).__init__(first=first, second=second,
                                         extensions = list(extensions), **params)
        self.pprint_args(['first', 'second'],[], infix_operator='|')

    def save(self, filename,  metadata={}, **data):
        raise Exception("A FileChoice cannot be used to save data.")

    def metadata(self, filename):
        self._loadpath(filename) # Ensures a valid extension

        try:    first_metadata = self.first.metadata(filename)
        except: first_metadata = {}
        try:     second_metadata = self.second.metadata(filename)
        except:  second_metadata = {}
        return dict(first_metadata, **second_metadata)

    def data(self, filename):
        self._loadpath(filename) # Ensures a valid extension

        try:     first_data = self.first.data(filename)
        except:  first_data = {}
        try:    second_data = self.second.data(filename)
        except: second_data = {}
        return dict(first_data, **second_data)

    def __repr__(self):
        return self._pprint(flat=True, annotate=False)

    def __str__(self):
        return self._pprint(annotate=False)


class CustomFile(FileType):
    """
    A customizable FileType that takes two functions as input and maps
    them to the loading interface for all FileTypes.
    """

    data_fn = param.Callable(doc='''
       A callable that takes a filename and returns a dictionary of
       data values''')

    metadata_fn = param.Callable(doc='''
        A callable that takes a filename and returns a dictionary of
        metadata values''')

    def __init__(self, data_fn=None, metadata_fn=None, **params):
        zipped = zip(['data_fn','metadata_fn'], [data_fn, metadata_fn])
        fn_dict = dict([(k,v) for (k,v) in zipped if (v is not None)])
        super(CustomFile, self).__init__(**dict(params, **fn_dict))

    def save(self, filename, data):
        raise NotImplementedError

    def data(self, filename):
        val = self.data_fn(filename)
        if not isinstance(val, dict):
            val = {self.data_key:val}
        return val

    def metadata(self, filename):
        val = self.metadata_fn(filename)
        if not isinstance(val, dict):
            raise Exception("The metadata callable must return a dictionary.")
        return val


class HVZFile(CustomFile):
    """
    FileType supporting the .hvz file format of the the HoloViews
    library (http://ioam.github.io/holoviews).

    Equivalent to the following CustomFile:

    CustomFile(metadata_fn=lambda f: Unpickler.key(f),
               data_fn = lambda f: {e: Unpickler.load(f, [e])
                                       for e in Unpickler.entries(f)})
    """

    def hvz_data_fn(f):
        from holoviews.core.io import Unpickler
        return {e: Unpickler.load(f, [e]) for e in Unpickler.entries(f)}

    def hvz_metadata_fn(f):
        from holoviews.core.io import Unpickler
        return Unpickler.key(f)

    data_fn = param.Callable(hvz_data_fn, doc="""
        By default loads all the entries in the .hvz file using
        Unpickler.load and returns them as a dictionary.""")

    metadata_fn = param.Callable(hvz_metadata_fn, doc="""
       Returns the key stored in the .hvz file as metadata using the
       Unpickler.key method.""")



class JSONFile(FileType):
    """
    It is assumed you won't store very large volumes of data as JSON.
    For this reason, the contents of JSON files are loaded as
    metadata.
    """
    extensions = param.List(default=['.json'], constant=True)

    def __init__(self, **params):
        super(JSONFile, self).__init__(**params)
        self.pprint_args(['hash_suffix'], [])

    def save(self, filename, metadata={}):
        jsonfile = open(self._savepath(filename),'wb')
        json.dump(metadata, jsonfile)

    def metadata(self, filename):
        jsonfile = open(self._loadpath(filename),'r')
        jsondata = json.load(jsonfile)
        jsonfile.close()
        return jsondata

    def data(self, filename):
        raise Exception("JSONFile only loads metadata")



class NumpyFile(FileType):
    """
    An npz file is the standard way to save Numpy arrays. This is a
    highly flexible FileType that supports most native Python objects
    including Numpy arrays.
    """

    extensions = param.List(default=['.npz'], constant=True)

    compress = param.Boolean(default=True, doc="""
      Whether or not the compressed npz format should be used.""")

    def __init__(self, **params):
        super(NumpyFile, self).__init__(**params)
        self.pprint_args(['hash_suffix'], ['compress']) # CHECK!

    def save(self, filename, metadata={}, **data):
        super(NumpyFile, self).save(filename, metadata, **data)
        savefn = numpy.savez_compressed if self.compress else numpy.savez
        savefn(self._savepath(filename), metadata=metadata, **data)

    def metadata(self, filename):
        npzfile = numpy.load(self._loadpath(filename))
        metadata = (npzfile['metadata'].tolist()
                    if 'metadata' in list(npzfile.keys()) else {})
        # Numpy load may return a Python dictionary.
        if not isinstance(npzfile, dict): npzfile.close()
        return metadata

    def data(self, filename):
        npzfile = numpy.load(self._loadpath(filename))
        keys = [k for k in npzfile.keys() if k != 'metadata']
        data = dict((k,npzfile[k]) for k in keys)

        # Is this a safe way to unpack objects?
        for (k,val) in data.items():
            if val.dtype.char == 'O' and val.shape == ():
                data[k] = val[()]

        if not isinstance(npzfile, dict):
            npzfile.close()
        return data


class ImageFile(FileType):
    """
    Image support - requires PIL or Pillow.
    """
    extensions = param.List(default=['.png', '.jpg'], constant=True)

    image_info = param.Dict(default={'mode':'mode', 'size':'size', 'format':'format'},
        doc='''Dictionary of the metadata to load. Each key is the
        name given to the metadata item and each value is the PIL
        Image attribute to return.''')

    data_mode = param.ObjectSelector(default='RGBA',
                                     objects=['L', 'RGB', 'RGBA', 'I','F'],
        doc='''Sets the mode of the mode of the Image object. Palette
        mode'P is not supported''')

    data_key = param.String(default='images', doc='''
       The name (key) given to the loaded image data.''')

    def __init__(self, **params):
        super(ImageFile, self).__init__(**params)
        self.pprint_args(['hash_suffix'],
                         ['data_key', 'data_mode', 'image_info'])

    def metadata(self, filename, **kwargs):
        image = Image.open(self._loadpath(filename))
        return dict((name, getattr(image,attr,None))
                    for (name, attr) in self.image_info.items())

    def save(self, filename, imdata, **data):
        """
        Data may be either a PIL Image object or a Numpy array.
        """
        if isinstance(imdata, numpy.ndarray):
            imdata = Image.fromarray(numpy.uint8(imdata))
        elif isinstance(imdata, Image.Image):
            imdata.save(self._savepath(filename))

    def data(self, filename):
        image = Image.open(self._loadpath(filename))
        data = image.convert(self.data_mode)
        return {self.data_key:data}


class MatplotlibFile(FileType):
    """
    Since version 1.0, Matplotlib figures support pickling. An mpkl
    file is simply a pickled matplotlib figure.
    """

    extensions = param.List(default=['.mpkl'], constant=True)

    def __init__(self, **params):
        super(MatplotlibFile, self).__init__(**params)
        self.pprint_args(['hash_suffix'], [])

    def save(self, filename, fig):
        pickle.dump(fig, open(self._savepath(filename),'wb'))

    def metadata(self, filename):
        pklfile = open(self._loadpath(filename),'r')
        fig = pickle.load(pklfile)
        metadata = {'dpi':fig.dip, 'size':fig.size}
        pklfile.close()
        return metadata

    def data(self, filename):
        pklfile = open(self._loadpath(filename),'r')
        fig = pickle.load(pklfile)
        pklfile.close()
        return {self.data_key:fig}



#
# Work in progress. Only SimpleOptimization is currently available.
#

import os, json, fnmatch
import param
from lancet.core import Arguments, Concatenate, CartesianProduct

class DynamicArgs(Arguments):
    """
    DynamicArgs are declarative specifications that specify a
    parameter space via a dynamic algorithmic process instead of using
    precomputed arguments. Unlike the static Args objects, new
    arguments can only be generated in response to some feedback from
    the processes that are executed. This type of dynamic feedback is
    a common feature for many algorithms such as hill climbing
    optimization, genetic algorithms, bisection search and other
    sophisticated optimization and search procedures.

    Like the Args objects, a DynamicArgs object is an iterator. On
    each iteration, one or more argument sets defining a collection of
    independent jobs are returned (these jobs should be possible to
    execute concurrently). Between iterations, the output_extractor
    function is used to extract the necessary information from the
    standard output streams of the previously executed jobs. This
    information is used to update the internal state of the
    DynamicArgs object can then generate more arguments to explore or
    terminate. All DynamicArgs classes need to declare the expected
    return format of the output_extractor function.
    """

    output_extractor = param.Callable(default=json.loads, doc="""
       The function that returns the relevant data from the standard
       output stream dumped to file in the streams subdirectory. This
       information must be retyrned in a format suitable for updating
       the specifier.. By default uses json.loads but pickle.loads
       could also be a valid option. The callable must take a string
       and return a Python object suitable for updating the specifier.""")


    def __init__(self, **params):
        super(DynamicArgs, self).__init__(**params)
        # Returned on next iteration and updated by _updated_state method
        self._next_val = self._initial_state(**params)
        # Trace of inputs from output_extractor and returned arguments
        self.trace = [(None, self._next_val)]

    def _update_state(self, data):
        """
        Determines the next desired point in the parameter space using
        the parsed data returned from the public update
        method. Returns the value that will next emitted on the next
        iteration using the list of dictionaries format for
        arguments. If the update fails or data is None, StopIteration
        should be supplied as the return value.
        """
        raise NotImplementedError


    def _initial_state(self, **kwargs):
        """
        Reset the the DynamicArgs object to its initial state and used
        to reset the object adter the iterator is exhausted. The
        return value is the initial argument to be returned by
        next().
        """
        raise NotImplementedError

    def __next__(self):
        if self._next_val is StopIteration:
            self._initial_state()
            raise StopIteration
        current_val =  self._next_val
        self._next_val = StopIteration
        return current_val

    next = __next__


    def update(self, tids, info):
        """
        Called to update the state of the iterator.  This methods
        receives the set of task ids from the previous set of tasks
        together with the launch information to allow the output
        values to be parsed using the output_extractor. This data is then
        used to determine the next desired point in the parameter
        space by calling the _update_state method.
        """
        outputs_dir = os.path.join(info['root_directory'], 'streams')
        pattern = '%s_*_tid_*{tid}.o.{tid}*' % info['batch_name']
        flist = os.listdir(outputs_dir)
        try:
            outputs = []
            for tid in tids:
                matches = fnmatch.filter(flist, pattern.format(tid=tid))
                if len(matches) != 1:
                    self.warning("No unique output file for tid %d" % tid)
                contents = open(os.path.join(outputs_dir, matches[0]),'r').read()
                outputs.append(self.output_extractor(contents))

            self._next_val = self._update_state(outputs)
            self.trace.append((outputs, self._next_val))
        except:
            self.warning("Cannot load required output files. Cannot continue.")
            self._next_val = StopIteration


    def show(self):
        """
        When dynamic, not all argument values may be available.
        """
        copied = self.copy()
        enumerated = [el for el in enumerate(copied)]
        for (group_ind, specs) in enumerated:
            if len(enumerated) > 1: print("Group %d" % group_ind)
            ordering = self.constant_keys + self.varying_keys
            # Ordered nicely by varying_keys definition.
            spec_lines = [', '.join(['%s=%s' % (k, s[k]) for k in ordering]) for s in specs]
            print('\n'.join(['%d: %s' % (i,l) for (i,l) in enumerate(spec_lines)]))

        print('Remaining arguments not available for %s' % self.__class__.__name__)


    def _trace_summary(self):
        """
        Summarizes the trace of values used to update the DynamicArgs
        and the arguments subsequently returned. May be used to
        implement the summary method.
        """
        for (i, (val, args)) in enumerate(self.trace):
            if args is StopIteration:
                info = "Terminated"
            else:
                pprint = ','.join('{' + ','.join('%s=%r' % (k,v)
                         for (k,v) in arg.items()) + '}' for arg in args)
                info = ("exploring arguments [%s]" % pprint )

            if i == 0: print("Step %d: Initially %s." % (i, info))
            else:      print("Step %d: %s after receiving input(s) %s." % (i, info.capitalize(), val))

    def __add__(self, other):
        """
        Concatenates two argument specifiers. See Concatenate and
        DynamicConcatenate documentation respectively.
        """
        if not other: return self
        dynamic = (isinstance(self, DynamicArgs),  isinstance(other, DynamicArgs))
        if dynamic == (True, True):
            raise Exception('Cannot concatenate two dynamic specifiers.')
        elif (True in dynamic):
            return DynamicConcatenate(self,other)
        else:
            return Concatenate(self,other)


    def __mul__(self, other):
        """
        Takes the cartesian product of two argument specifiers. See
        CartesianProduct and DynamicCartesianProduct documentation.
        """
        if not other: return []
        dynamic = (isinstance(self, DynamicArgs),  isinstance(other, DynamicArgs))
        if dynamic == (True, True):
            raise Exception('Cannot take Cartesian product two dynamic specifiers.')
        elif (True in dynamic):
            return DynamicCartesianProduct(self, other)
        else:
            return CartesianProduct(self, other)


    def __len__(self):
        """
        Many DynamicArgs won't have a length that can be
        precomputed. Most DynamicArgs objects will have an iteration
        limit to guarantee eventual termination. If so, the maximum
        possible number of arguments that could be generated should be
        returned.
        """
        raise NotImplementedError



class SimpleGradientDescent(DynamicArgs):
    """
    Very simple gradient descent optimizer designed to illustrate how
    Dynamic Args may be implemented. This class has been deliberately
    kept simple to clearly illustrate how Dynamic Args may be
    implemented. A more practical example would likely probably make
    use of mature, third party optimization libraries (such as the
    routines offered in scipy.optimize).

    This particular algorithm greedily minimizes an output value via
    greedy gradient descent. The local parameter space is explored by
    examining the change in output value when an increment or
    decrement of 'stepsize' is made in the parameter space, centered
    around the current position. The initial parameter is initialized
    with the 'start' value and the optimization process terminates
    when either a local minima/maxima has been found or when
    'max_steps' is reached.

    The 'output_extractor' function is expected to return a single
    scalar number to drive the gradient descent algorithm forwards.
    """

    key = param.String(constant=True, doc="""
        The name of the argument that will be optimized in a greedy fashion.""")

    start = param.Number(default=0.0, constant=True, doc="""
        The starting argument value for the gradient ascent or descent""")

    stepsize = param.Number(default=1.0, constant=True, doc="""
        The size of the steps taken in parameter space.""")

    max_steps=param.Integer(default=100, constant=True, doc="""
        Once max_steps is reached, the optimization terminates.""")

    def __init__(self, key, **params):
        super(SimpleGradientDescent, self).__init__(key=key, **params)
        self.pprint_args(['key', 'start', 'stepsize'],[])
        self._termination_info = None

    def _initial_state(self, **kwargs):
        self._steps_complete = 0
        self._best_val = float('inf')
        self._arg = self.start
        return [{self.key:self.start+1}, {self.key:self.start-1}]

    def _update_state(self, vals):
        """
        Takes as input a list or tuple of two elements. First the
        value returned by incrementing by 'stepsize' followed by the
        value returned after a 'stepsize' decrement.
        """
        self._steps_complete += 1
        if self._steps_complete == self.max_steps:
            self._termination_info = (False, self._best_val, self._arg)
            return StopIteration

        arg_inc, arg_dec = vals
        best_val = min(arg_inc, arg_dec, self._best_val)
        if best_val == self._best_val:
            self._termination_info = (True, best_val, self._arg)
            return StopIteration

        self._arg += self.stepsize if (arg_dec > arg_inc) else -self.stepsize
        self._best_val= best_val
        return [{self.key:self._arg+self.stepsize},
                {self.key:self._arg-self.stepsize}]

    @property
    def constant_keys(self):  return []
    @property
    def constant_items(self): return []
    @property
    def varying_keys(self):   return [self.key]

    def summary(self):
        print('Varying Keys: %r' % self.key)
        print('Maximum steps allowed: %d' % self.max_steps)
        self._trace_summary()
        (val, arg) = (self.trace[-1])
        if self._termination_info:
            (success, best_val, arg) = self._termination_info
            condition =  'Successfully converged.' if success else 'Maximum step limit reached.'
            print("%s Minimum value of %r at %s=%r." % (condition, best_val, self.key, arg))

    def __len__(self):
        return 2*self.max_steps # Each step specifies 2 concurrent jobs



#=========================#
# Experimental code (WIP) #
#=========================#


class DynamicConcatenate(DynamicArgs):
    def __init__(self, first, second):
        self.first = first
        self.second = second
        super(Concatenate, self).__init__(dynamic=True)

        self._exhausted = False
        self._first_sent = False
        self._first_cached = None
        self._second_cached = None
        if not isinstance(first,DynamicArgs):
            self._first_cached = next(first.copy())
        if not isinstance(second,DynamicArgs):
            self._second_cached = next(second.copy())
        self.pprint_args(['first', 'second'],[], infix_operator='+')


    def constant_keys(self):
        return list(set(self.first.constant_keys) | set(self.second.constant_keys))


    def varying_keys(self):
        return list(set(self.first.varying_keys) | set(self.second.varying_keys))


    def update(self, tids, info):
        if (self.isinstance(self.first,DynamicArgs) and not self._exhausted):
            self.first.update(tids, info)
        elif (self.isinstance(self.second,DynamicArgs) and self._first_sent):
            self.second.update(tids, info)

    def __next__(self):
        if self._first_cached is None:
            try:  return next(self.first)
            except StopIteration:
                self._exhausted = True
                return self._second_cached
        else:
            if not self._first_sent:
                self._first_sent = True
                return self._first_cached
            else:
                return  next(self.second)

    next = __next__


class DynamicCartesianProduct(DynamicArgs):

    def __init__(self, first, second):

        self.first = first
        self.second = second

        overlap = set(self.first.varying_keys) &  set(self.second.varying_keys)
        assert overlap == set(), 'Sets of keys cannot overlap between argument specifiers in cartesian product.'

        super(CartesianProduct, self).__init__(dynamic=True)

        self._first_cached = None
        self._second_cached = None
        if not isinstance(first,DynamicArgs):
            self._first_cached = next(first.copy())
        if not isinstance(second,DynamicArgs):
            self._second_cached = next(second.copy())

        self.pprint_args(['first', 'second'],[], infix_operator='*')

    def constant_keys(self):
        return list(set(self.first.constant_keys) | set(self.second.constant_keys))


    def varying_keys(self):
        return list(set(self.first.varying_keys) | set(self.second.varying_keys))


    def update(self, tids, info):
        if self.isinstance(self.first,DynamicArgs):  self.first.update(tids, info)
        if self.isinstance(self.second,DynamicArgs): self.second.update(tids, info)


    def __next__(self):
        if self._first_cached is None:
            first_spec = next(self.first)
            return self._cartesian_product(first_spec, self._second_cached)
        else:
            second_spec = next(self.second)
            return self._cartesian_product(self._first_cached, second_spec)
    next = __next__



#
# Lancet core
#

import os, itertools, copy
import re, glob, string
import json

import param

try:
    import numpy as np
    np_ftypes = np.sctypes['float']
except:
    np, np_ftypes = None, []

try:    from pandas import DataFrame
except: DataFrame = None # pyflakes:ignore (try/except import)

try: from holoviews import Table
except: Table = None     # pyflakes:ignore (try/except import)

from collections import defaultdict, OrderedDict

float_types = [float] + np_ftypes
def identityfn(x): return x
def fp_repr(x):    return str(x) if (type(x) in float_types) else repr(x)

def set_fp_precision(value):
    """
    Function to set the floating precision across lancet.
    """
    Arguments.set_default('fp_precision', value)

#=====================#
# Argument Specifiers #
#=====================#

class PrettyPrinted(object):
    """
    A mixin class for generating pretty-printed representations.
    """

    def pprint_args(self, pos_args, keyword_args, infix_operator=None, extra_params={}):
        """
        Method to define the positional arguments and keyword order
        for pretty printing.
        """
        if infix_operator and not (len(pos_args)==2 and keyword_args==[]):
            raise Exception('Infix format requires exactly two'
                            ' positional arguments and no keywords')
        (kwargs,_,_,_) = self._pprint_args
        self._pprint_args = (keyword_args + kwargs, pos_args, infix_operator, extra_params)

    def _pprint(self, cycle=False, flat=False, annotate=False, onlychanged=True, level=1, tab = '   '):
        """
        Pretty printer that prints only the modified keywords and
        generates flat representations (for repr) and optionally
        annotates the top of the repr with a comment.
        """
        (kwargs, pos_args, infix_operator, extra_params) = self._pprint_args
        (br, indent)  = ('' if flat else '\n', '' if flat else tab * level)
        prettify = lambda x: isinstance(x, PrettyPrinted) and not flat
        pretty = lambda x: x._pprint(flat=flat, level=level+1) if prettify(x) else repr(x)

        params = dict(self.get_param_values())
        show_lexsort = getattr(self, '_lexorder', None) is not None
        modified = [k for (k,v) in self.get_param_values(onlychanged=onlychanged)]
        pkwargs = [(k, params[k])  for k in kwargs if (k in modified)] + list(extra_params.items())
        arg_list = [(k,params[k]) for k in pos_args] + pkwargs

        lines = []
        if annotate: # Optional annotating comment
            len_ckeys, len_vkeys = len(self.constant_keys), len(self.varying_keys)
            info_triple = (len(self),
                           ', %d constant key(s)' % len_ckeys if len_ckeys else '',
                           ', %d varying key(s)'  % len_vkeys if len_vkeys else '')
            annotation = '# == %d items%s%s ==\n' % info_triple
            lines = [annotation]

        if show_lexsort: lines.append('(')
        if cycle:
            lines.append('%s(...)' % self.__class__.__name__)
        elif infix_operator:
            level = level - 1
            triple = (pretty(params[pos_args[0]]), infix_operator, pretty(params[pos_args[1]]))
            lines.append('%s %s %s' % triple)
        else:
            lines.append('%s(' % self.__class__.__name__)
            for (k,v) in arg_list:
                lines.append('%s%s=%s' %  (br+indent, k, pretty(v)))
                lines.append(',')
            lines = lines[:-1] +[br+(tab*(level-1))+')'] # Remove trailing comma

        if show_lexsort:
            lines.append(').lexsort(%s)' % ', '.join(repr(el) for el in self._lexorder))
        return ''.join(lines)

    def __repr__(self):
        return self._pprint(flat=True, onlychanged=False)

    def __str__(self):
        return self._pprint()



class Arguments(PrettyPrinted, param.Parameterized):
    """
    The abstract, base class that defines the core interface and
    methods for all members of the Arguments family of classes,
    including either the simple, static members of Args below, or the
    sophisticated parameter exploration algorithms subclassing from
    DynamicArgs defined in dynamic.py.

    The Args subclass may be used directly and forms the root of one
    family of classes that have statically defined or precomputed
    argument sets (defined below). The second subfamily are the
    DynamicArgs, designed to allow more sophisticated, online
    parameter space exploration techniques such as hill climbing,
    bisection search, genetic algorithms and so on.
    """

    fp_precision = param.Integer(default=4, constant=True, doc='''
         The floating point precision to use for floating point
         values.  Unlike other basic Python types, floats need care
         with their representation as you only want to display up to
         the precision actually specified. A floating point precision
         of 0 casts number to integers before representing them.''')

    def __init__(self, **params):
        self._pprint_args = ([],[],None,{})
        self.pprint_args([],['fp_precision', 'dynamic'])
        super(Arguments,self).__init__(**params)
        # Some types cannot be sorted easily (e.g. numpy arrays)
        self.unsortable_keys = []

    def __iter__(self): return self

    def __contains__(self, value):
        return value in (self.constant_keys + self.varying_keys)

    @classmethod
    def spec_formatter(cls, spec):
        " Formats the elements of an argument set appropriately"
        return type(spec)((k, str(v)) for (k,v) in spec.items())

    @property
    def constant_keys(self):
        """
        Returns the list of parameter names whose values are constant
        as the argument specifier is iterated.  Note that the union of
        constant and varying_keys should partition the entire set of
        keys in the case where there are no unsortable keys.
        """
        raise NotImplementedError

    @property
    def constant_items(self):
        """
        Returns the set of constant items as a list of tuples. This
        allows easy conversion to dictionary format. Note, the items
        should be supplied in the same key ordering as for
        constant_keys for consistency.
        """
        raise NotImplementedError

    @property
    def varying_keys(self):
        """
        Returns the list of parameters whose values vary as the
        argument specifier is iterated.  Whenever it is possible, keys
        should be sorted from those slowest to faster varying and
        sorted alphanumerically within groups that vary at the same
        rate.
        """
        raise NotImplementedError

    def round_floats(self, specs, fp_precision):
        _round_float = lambda v, fp: np.round(v, fp) if (type(v) in np_ftypes) else round(v, fp)
        _round = (lambda v, fp: int(v)) if fp_precision==0  else _round_float
        return (dict((k, _round(v, fp_precision) if (type(v) in float_types) else v)
                     for (k,v) in spec.items()) for spec in specs)

    def __next__(self):
        """
        Called to get a list of specifications: dictionaries with
        parameter name keys and string values.
        """
        raise StopIteration

    next = __next__

    def copy(self):
        """
        Convenience method to avoid using the specifier without
        exhausting it.
        """
        return copy.copy(self)

    def _collect_by_key(self,specs):
        """
        Returns a dictionary like object with the lists of values
        collapsed by their respective key. Useful to find varying vs
        constant keys and to find how fast keys vary.
        """
        # Collect (key, value) tuples as list of lists, flatten with chain
        allkeys = itertools.chain.from_iterable(
            [[(k, run[k]) for k in run] for run in specs])
        collection = defaultdict(list)
        for (k,v) in allkeys: collection[k].append(v)
        return collection

    def _operator(self, operator, other):
        identities = [isinstance(el, Identity) for el in [self, other]]
        if  not any(identities): return operator(self,other)
        if all(identities):      return Identity()
        elif identities[1]:      return self
        else:                    return other

    def __add__(self, other):
        """
        Concatenates two argument specifiers.
        """
        return self._operator(Concatenate, other)

    def __mul__(self, other):
        """
        Takes the Cartesian product of two argument specifiers.
        """
        return self._operator(CartesianProduct, other)

    def _cartesian_product(self, first_specs, second_specs):
        """
        Takes the Cartesian product of the specifications. Result will
        contain N specifications where N = len(first_specs) *
        len(second_specs) and keys are merged.
        Example: [{'a':1},{'b':2}] * [{'c':3},{'d':4}] =
        [{'a':1,'c':3},{'a':1,'d':4},{'b':2,'c':3},{'b':2,'d':4}]
        """
        return  [ dict(zip(
                          list(s1.keys()) + list(s2.keys()),
                          list(s1.values()) + list(s2.values())
                      ))
                 for s1 in first_specs for s2 in second_specs ]

    def summary(self):
        """
        A succinct summary of the argument specifier. Unlike the repr,
        a summary does not have to be complete but must supply the
        most relevant information about the object to the user.
        """
        print("Items: %s" % len(self))
        varying_keys = ', '.join('%r' % k for k in self.varying_keys)
        print("Varying Keys: %s" % varying_keys)
        items = ', '.join(['%s=%r' % (k,v)
                           for (k,v) in self.constant_items])
        if self.constant_items:
            print("Constant Items: %s" % items)


class Identity(Arguments):
    """
    The identity element for any Arguments object 'args' under the *
    operator (CartesianProduct) and + operator (Concatenate). The
    following identities hold:

    args is (Identity() * args)
    args is (args * Identity())

    args is (Identity() + args)
    args is (args + Identity())

    Note that the empty Args() object can also fulfill the role of
    Identity under the addition operator.
    """

    fp_precision = param.Integer(default=None, allow_None=True,
                                 precedence=(-1), constant=True, doc='''
       fp_precision is disabled as Identity() never contains any
       arguments.''')

    def __eq__(self, other): return isinstance(other, Identity)
    def __repr__(self): return "Identity()"
    def __str__(self): return repr(self)
    def __nonzero__(self): raise ValueError("The boolean value of Identity is undefined")
    def __bool__(self): raise ValueError("The boolean value of Identity is undefined")


class Args(Arguments):
    """
    An Arguments class that supports statically specified or
    precomputed argument sets. It may be used directly to specify
    argument values but also forms the base class for a family of more
    specific static Argument classes. Each subclass is less flexible
    and general but allows arguments to be easily and succinctly
    specified. For instance, the Range subclass allows parameter
    ranges to be easily declared.

    The constructor of Args accepts argument definitions in two
    different formats. The keyword format allows constant arguments to
    be specified directly and easily. For instance:

    >>> v1 = Args(a=2, b=3)
    >>> v1
    Args(fp_precision=4,a=2,b=3)

    The alternative input format takes an explicit list of the
    argument specifications:

    >>> v2 = Args([{'a':2, 'b':3}]) # Equivalent behaviour to above
    >>> v1.specs  == v2.specs
    True

    This latter format is completely flexible and general, allowing
    any arbitrary list of arguments to be specified as desired. This
    is not generally recommended however as the structure of a
    parameter space is often expressed more clearly by composing
    together simpler, more succinct Args objects with the
    CartesianProduct (*) or Concatenation (+) operators.
    """

    specs = param.List(default=[], constant=True, doc='''
          The static list of specifications (ie. dictionaries) to be
          returned by the specifier. Float values are rounded
          according to fp_precision.''')

    def __init__(self, specs=None, fp_precision=None, **params):
        if fp_precision is None: fp_precision = Arguments.fp_precision
        raw_specs, params, explicit = self._build_specs(specs, params, fp_precision)
        super(Args, self).__init__(fp_precision=fp_precision, specs=raw_specs, **params)

        self._lexorder = None
        if explicit:
            self.pprint_args(['specs'],[])
        else: # Present in kwarg format
            self.pprint_args([], self.constant_keys, None,
                             OrderedDict(sorted(self.constant_items)))

    def _build_specs(self, specs, kwargs, fp_precision):
        """
        Returns the specs, the remaining kwargs and whether or not the
        constructor was called with kwarg or explicit specs.
        """
        if specs is None:
            overrides = param.ParamOverrides(self, kwargs,
                                             allow_extra_keywords=True)
            extra_kwargs = overrides.extra_keywords()
            kwargs = dict([(k,v) for (k,v) in kwargs.items()
                           if k not in extra_kwargs])
            rounded_specs = list(self.round_floats([extra_kwargs],
                                                   fp_precision))

            if extra_kwargs=={}: return [], kwargs, True
            else:                return rounded_specs, kwargs, False

        return list(self.round_floats(specs, fp_precision)), kwargs, True

    def __iter__(self):
        self._exhausted = False
        return self

    def __next__(self):
        if self._exhausted:
            raise StopIteration
        else:
            self._exhausted=True
            return self.specs

    next = __next__

    def _unique(self, sequence, idfun=repr):
        """
        Note: repr() must be implemented properly on all objects. This
        is implicitly assumed by Lancet when Python objects need to be
        formatted to string representation.
        """
        seen = {}
        return [seen.setdefault(idfun(e),e) for e in sequence
                if idfun(e) not in seen]

    def show(self, exclude=[]):
        """
        Convenience method to inspect the available argument values in
        human-readable format. The ordering of keys is determined by
        how quickly they vary.

        The exclude list allows specific keys to be excluded for
        readability (e.g. to hide long, absolute filenames).
        """
        ordering = self.constant_keys + self.varying_keys
        spec_lines = [', '.join(['%s=%s' % (k, s[k]) for k in ordering
                                 if (k in s) and (k not in exclude)])
                      for s in self.specs]
        print('\n'.join(['%d: %s' % (i,l) for (i,l) in enumerate(spec_lines)]))


    def lexsort(self, *order):
        """
        The lexical sort order is specified by a list of string
        arguments. Each string is a key name prefixed by '+' or '-'
        for ascending and descending sort respectively. If the key is
        not found in the operand's set of varying keys, it is ignored.
        """
        if order == []:
            raise Exception("Please specify the keys for sorting, use"
                            "'+' prefix for ascending,"
                            "'-' for descending.)")

        if not set(el[1:] for el in order).issubset(set(self.varying_keys)):
            raise Exception("Key(s) specified not in the set of varying keys.")

        sorted_args = copy.deepcopy(self)
        specs_param = sorted_args.params('specs')
        specs_param.constant = False
        sorted_args.specs = self._lexsorted_specs(order)
        specs_param.constant = True
        sorted_args._lexorder = order
        return sorted_args

    def _lexsorted_specs(self, order):
        """
        A lexsort is specified using normal key string prefixed by '+'
        (for ascending) or '-' for (for descending).

        Note that in Python 2, if a key is missing, None is returned
        (smallest Python value). In Python 3, an Exception will be
        raised regarding comparison of heterogenous types.
        """
        specs = self.specs[:]
        if not all(el[0] in ['+', '-'] for el in order):
            raise Exception("Please specify the keys for sorting, use"
                            "'+' prefix for ascending,"
                            "'-' for descending.)")

        sort_cycles = [(el[1:], True if el[0]=='+' else False)
                       for el in reversed(order)
                       if el[1:] in self.varying_keys]

        for (key, ascending) in sort_cycles:
            specs = sorted(specs, key=lambda s: s.get(key, None),
                           reverse=(not ascending))
        return specs

    @property
    def constant_keys(self):
        collection = self._collect_by_key(self.specs)
        return [k for k in sorted(collection) if
                (len(self._unique(collection[k])) == 1)]

    @property
    def constant_items(self):
        collection = self._collect_by_key(self.specs)
        return [(k,collection[k][0]) for k in self.constant_keys]

    @property
    def varying_keys(self):
        collection = self._collect_by_key(self.specs)
        constant_set = set(self.constant_keys)
        unordered_varying = set(collection.keys()).difference(constant_set)
        # Finding out how fast keys are varying
        grouplens = [(len([len(list(y)) for (_,y)
                           in itertools.groupby(collection[k])]),k)
                     for k in collection
                     if (k not in self.unsortable_keys)]
        varying_counts = [(n,k) for (n,k) in sorted(grouplens) if (k in unordered_varying)]
        # Grouping keys with common frequency alphanumerically (desired behaviour).
        ddict = defaultdict(list)
        for (n,k) in varying_counts: ddict[n].append(k)
        alphagroups = [sorted(ddict[k]) for k in sorted(ddict)]
        return [el for group in alphagroups for el in group] + sorted(self.unsortable_keys)

    @property
    def dframe(self):
        return DataFrame(self.specs) if DataFrame else "Pandas not available"

    @property
    def table(self):
        if not Table:
            return "HoloViews Table not available"
        keys =  self.varying_keys + self.constant_keys
        items = [(tuple([spec[k] for k  in keys]),()) for spec in self.specs]
        return Table(items, key_dimensions=keys, value_dimensions=[])


    def __len__(self): return len(self.specs)



class Concatenate(Args):
    """
    Concatenate is the sequential composition of two specifiers. The
    specifier created by the compositon (firsts + second) generates
    the arguments in first followed by the arguments in second.
    """

    first = param.ClassSelector(default=None, class_=Args, allow_None=True,
       constant=True, doc='''The first specifier in the concatenation.''')

    second = param.ClassSelector(default=None, class_=Args, allow_None=True,
       constant=True, doc='''The second specifier in the concatenation.''')

    def __init__(self, first, second):

        max_precision = max(first.fp_precision, second.fp_precision)
        specs = first.specs + second.specs
        super(Concatenate, self).__init__(specs, fp_precision=max_precision,
                                                first=first, second=second)
        self.pprint_args(['first', 'second'],[], infix_operator='+')


class CartesianProduct(Args):
    """
    CartesianProduct is the Cartesian product of two specifiers. The
    specifier created by the compositon (firsts * second) generates
    the cartesian produce of the arguments in first followed by the
    arguments in second. Note that len(first * second) =
    len(first)*len(second)
    """

    first = param.ClassSelector(default=None, class_=Args, allow_None=True,
       constant=True, doc='''The first specifier in the Cartesian product.''')

    second = param.ClassSelector(default=None, class_=Args, allow_None=True,
       constant=True, doc='''The second specifier in the Cartesian product.''')

    def __init__(self, first, second):

        max_precision = max(first.fp_precision, second.fp_precision)
        specs = self._cartesian_product(first.specs, second.specs)

        overlap = (set(first.varying_keys + first.constant_keys)
                   &  set(second.varying_keys + second.constant_keys))
        assert overlap == set(), ('Sets of keys cannot overlap'
                                  'between argument specifiers'
                                  'in cartesian product.')
        super(CartesianProduct, self).__init__(specs, fp_precision=max_precision,
                                               first=first, second=second)
        self.pprint_args(['first', 'second'],[], infix_operator='*')


class Range(Args):
    """
    Range generates an argument from a numerically interpolated range
    which is linear by default. An optional function can be specified
    to sample a numeric range with regular intervals.
    """

    key = param.String(default='', constant=True, doc='''
         The key assigned to the values computed over the numeric range.''')

    start_value =  param.Number(default=None, allow_None=True, constant=True,
        doc='''The starting numeric value of the range.''')

    end_value = param.Number(default=None, allow_None=True, constant=True,
       doc='''The ending numeric value of the range (inclusive).''')

    steps = param.Integer(default=2, constant=True, bounds=(1,None),
       doc='''The number of steps to interpolate over. Default is 2
         which returns the start and end values without interpolation.''')

    # Can't this be a lambda?
    mapfn = param.Callable(default=identityfn, constant=True, doc='''
         The function to be mapped across the linear range. The
         identity function is used by by default''')

    def __init__(self, key, start_value, end_value, steps=2, mapfn=identityfn, **params):

        values = self.linspace(start_value, end_value, steps)
        specs = [{key:mapfn(val)} for val in values ]

        super(Range, self).__init__(specs, key=key, start_value=start_value,
                                         end_value=end_value, steps=steps,
                                         mapfn=mapfn, **params)
        self.pprint_args(['key', 'start_value'], ['end_value', 'steps'])

    def linspace(self, start, stop, n):
        """ Simple replacement for numpy linspace"""
        if n == 1: return [start]
        L = [0.0] * n
        nm1 = n - 1
        nm1inv = 1.0 / nm1
        for i in range(n):
            L[i] = nm1inv * (start*(nm1 - i) + stop*i)
        return L

class List(Args):
    """
    An argument specifier that takes its values from a given list.
    """

    values = param.List(default=[], constant=True, doc='''
         The list values that are to be returned by the specifier''')

    key = param.String(default='default', constant=True, doc='''
         The key assigned to the elements of the supplied list.''')

    def __init__(self, key, values, **params):
        specs = [{key:val} for val in values]
        super(List, self).__init__(specs, key=key, values=values, **params)
        self.pprint_args(['key', 'values'], [])


class Log(Args):
    """
    Specifier that loads arguments from a log file in task id (tid)
    order.  This wrapper class allows a concise representation of file
    logs with the option of adding the task id to the loaded
    specifications.

    For full control over the arguments, you can use this class to
    create a fully specified Args object as follows:

    Args(Log.extract_log(<log_file>).values()),
    """

    log_path = param.String(default=None, allow_None=True, constant=True,
         doc='''The relative or absolute path to the log file. If a
              relative path is given, the absolute path is computed
              relative to os.getcwd().''')

    tid_key = param.String(default='tid', constant=True, allow_None=True,
         doc='''If not None, the key given to the tid values included
               in the loaded specifications. If None, the tid number
               is ignored.''')

    @staticmethod
    def extract_log(log_path, dict_type=dict):
        """
        Parses the log file generated by a launcher and returns
        dictionary with tid keys and specification values.

        Ordering can be maintained by setting dict_type to the
        appropriate constructor (i.e. OrderedDict). Keys are converted
        from unicode to strings for kwarg use.
        """
        log_path = (log_path if os.path.isfile(log_path)
                    else os.path.join(os.getcwd(), log_path))
        with open(log_path,'r') as log:
            splits = (line.split() for line in log)
            uzipped = ((int(split[0]), json.loads(" ".join(split[1:]))) for split in splits)
            szipped = [(i, dict((str(k),v) for (k,v) in d.items())) for (i,d) in uzipped]
        return dict_type(szipped)

    @staticmethod
    def write_log(log_path, data, allow_append=True):
        """
        Writes the supplied specifications to the log path. The data
        may be supplied as either as a an Args or as a list of
        dictionaries.

        By default, specifications will be appropriately appended to
        an existing log file. This can be disabled by setting
        allow_append to False.
        """
        append = os.path.isfile(log_path)
        islist = isinstance(data, list)

        if append and not allow_append:
            raise Exception('Appending has been disabled'
                            ' and file %s exists' % log_path)

        if not (islist or isinstance(data, Args)):
            raise Exception('Can only write Args objects or dictionary'
                            ' lists to log file.')

        specs = data if islist else data.specs
        if not all(isinstance(el,dict) for el in specs):
            raise Exception('List elements must be dictionaries.')

        log_file = open(log_path, 'r+') if append else open(log_path, 'w')
        start = int(log_file.readlines()[-1].split()[0])+1 if append else 0
        ascending_indices = range(start, start+len(data))

        log_str = '\n'.join(['%d %s' % (tid, json.dumps(el))
                             for (tid, el) in zip(ascending_indices,specs)])
        log_file.write("\n"+log_str if append else log_str)
        log_file.close()

    def __init__(self, log_path, tid_key='tid', **params):

        log_items = sorted(Log.extract_log(log_path).items())

        if tid_key is None:
            log_specs = [spec for (_, spec) in log_items]
        else:
            log_specs = [dict(list(spec.items())+[(tid_key,idx)])
                         for (idx, spec) in log_items]

        super(Log, self).__init__(log_specs,
                                  log_path=log_path,
                                  tid_key=tid_key,
                                  **params)
        self.pprint_args(['log_path'], ['tid_key'])



class FilePattern(Args):
    """
    A FilePattern specifier allows files to be matched and information
    encoded in filenames to be extracted via an extended form of
    globbing. This object may be used to specify filename arguments to
    CommandTemplates when launching jobs but it also very useful for
    collating files for analysis.

    For instance, you can find the absolute filenames of all npz files
    in a 'data' subdirectory (relative to the root) that start with
    'timeseries' using the pattern 'data/timeseries*.npz'.

    In addition to globbing supported by the glob module, patterns can
    extract metadata encoded in filenames using a subset of the Python
    format specification syntax. To illustrate, you can use
    'data/timeseries-{date}.npz' to record the date strings associated
    with matched files. Note that a particular named fields can only
    be used in a particular pattern once.

    By default metadata is extracted as strings but format types are
    supported in the usual manner
    eg. 'data/timeseries-{day:d}-{month:d}.npz' will extract the day
    and month from the filename as integer values. Only field names
    and types are recognised with other format specification syntax
    ignored. Type codes supported: 'd', 'b', 'o', 'x', 'e','E','f',
    'F','g', 'G', 'n' (if ommited, result is a string by default).

    Note that ordering is determined via ascending alphanumeric sort
    and that actual filenames should not include any globbing
    characters, namely: '?','*','[' and ']' (general good practice for
    filenames anyway).
    """

    key = param.String(default=None, allow_None=True, constant=True,
       doc='''The key name given to the matched file path strings.''')

    pattern = param.String(default=None, allow_None=True, constant=True,
       doc='''The pattern files are to be searched against.''')

    root = param.String(default=None, allow_None=True, constant=True,
       doc='''The root directory from which patterns are to be loaded.
       The root is set relative to os.getcwd().''')

    @classmethod
    def directory(cls, directory, root=None, extension=None, **kwargs):
        """
        Load all the files in a given directory selecting only files
        with the given extension if specified. The given kwargs are
        passed through to the normal constructor.
        """
        root = os.getcwd() if root is None else root
        suffix = '' if extension is None else '.' + extension.rsplit('.')[-1]
        pattern = directory + os.sep + '*' + suffix
        key = os.path.join(root, directory,'*').rsplit(os.sep)[-2]
        format_parse = list(string.Formatter().parse(key))
        if not all([el is None for el in zip(*format_parse)[1]]):
            raise Exception('Directory cannot contain format field specifications')
        return cls(key, pattern, root, **kwargs)

    def __init__(self, key, pattern, root=None, **params):
        root = os.getcwd() if root is None else root
        specs = self._load_expansion(key, root, pattern)
        super(FilePattern, self).__init__(specs, key=key, pattern=pattern,
                                          root=root, **params)
        self.pprint_args(['key', 'pattern'], ['root'])

    def fields(self):
        """
        Return the fields specified in the pattern using Python's
        formatting mini-language.
        """
        parse = list(string.Formatter().parse(self.pattern))
        return [f for f in zip(*parse)[1] if f is not None]

    def _load_expansion(self, key, root, pattern):
        """
        Loads the files that match the given pattern.
        """
        path_pattern = os.path.join(root, pattern)
        expanded_paths = self._expand_pattern(path_pattern)

        specs=[]
        for (path, tags) in expanded_paths:
            filelist = [os.path.join(path,f) for f in os.listdir(path)] if os.path.isdir(path) else [path]
            for filepath in filelist:
                specs.append(dict(tags,**{key:os.path.abspath(filepath)}))

        return sorted(specs, key=lambda s: s[key])

    def _expand_pattern(self, pattern):
        """
        From the pattern decomposition, finds the absolute paths
        matching the pattern.
        """
        (globpattern, regexp, fields, types) = self._decompose_pattern(pattern)
        filelist = glob.glob(globpattern)
        expansion = []

        for fname in filelist:
            if fields == []:
                expansion.append((fname, {}))
                continue
            match = re.match(regexp, fname)
            if match is None: continue
            match_items = match.groupdict().items()
            tags = dict((k,types.get(k, str)(v)) for (k,v) in match_items)
            expansion.append((fname, tags))

        return expansion

    def _decompose_pattern(self, pattern):
        """
        Given a path pattern with format declaration, generates a
        four-tuple (glob_pattern, regexp pattern, fields, type map)
        """
        sep = '~lancet~sep~'
        float_codes = ['e','E','f', 'F','g', 'G', 'n']
        typecodes = dict([(k,float) for k in float_codes]
                         + [('b',bin), ('d',int), ('o',oct), ('x',hex)])
        parse = list(string.Formatter().parse(pattern))
        text, fields, codes, _ = zip(*parse)

        # Finding the field types from format string
        types = []
        for (field, code) in zip(fields, codes):
            if code in ['', None]: continue
            constructor =  typecodes.get(code[-1], None)
            if constructor: types += [(field, constructor)]

        stars =  ['' if not f else '*' for f in fields]
        globpat = ''.join(text+star for (text,star) in zip(text,stars))

        refields = ['' if not f else sep+('(?P<%s>.*?)'% f)+sep for f in fields]
        parts = ''.join(text+group for (text,group) in zip(text, refields)).split(sep)
        for i in range(0, len(parts), 2): parts[i] = re.escape(parts[i])

        regexp_pattern = ''.join(parts).replace('\\*','.*')
        fields = list(f for f in fields if f)
        return globpat, regexp_pattern , fields, dict(types)


# Importing from filetypes requires PrettyPrinted to be defined first
from lancet.filetypes import FileType

class FileInfo(Args):
    """
    Loads metadata from a set of filenames. For instance, you can load
    metadata associated with a series of image files given by a
    FilePattern. Unlike other explicit instances of Args, this object
    extends the values of an existing Args object. Once you have
    loaded the metadata, FileInfo allows you to load the file data
    into a pandas DataFrame or a HoloViews Table.
    """

    source = param.ClassSelector(class_ = Args, doc='''
        The argument specifier that supplies the file paths.''')

    filetype = param.ClassSelector(constant=True, class_= FileType, doc='''
        A FileType object to be applied to each file path.''')

    key = param.String(constant=True, doc='''
       The key used to find the file paths for inspection.''')

    ignore = param.List(default=[], constant=True, doc='''
       Metadata keys that are to be explicitly ignored. ''')

    def __init__(self, source, key, filetype, ignore = [], **params):
        specs = self._info(source, key, filetype, ignore)
        super(FileInfo, self).__init__(specs,
                                       source = source,
                                       filetype = filetype,
                                       key = key,
                                       ignore=ignore,
                                       **params)
        self.pprint_args(['source', 'key', 'filetype'], ['ignore'])


    @classmethod
    def from_pattern(cls, pattern, filetype=None, key='filename', root=None, ignore=[]):
        """
        Convenience method to directly chain a pattern processed by
        FilePattern into a FileInfo instance.

        Note that if a default filetype has been set on FileInfo, the
        filetype argument may be omitted.
        """
        filepattern = FilePattern(key, pattern, root=root)
        if FileInfo.filetype and filetype is None:
            filetype = FileInfo.filetype
        elif filetype is None:
            raise Exception("The filetype argument must be supplied unless "
                            "an appropriate default has been specified as "
                            "FileInfo.filetype")
        return FileInfo(filepattern, key, filetype, ignore=ignore)


    @property
    def table(self):
        """
        Return an ndmapping of the loaded data using the filenames as
        values and the remaining data as the keys.
        """
        all_dimension_labels = self.constant_keys + self.varying_keys
        dimension_labels = [d for d in all_dimension_labels if d != self.key]

        if dimension_labels == []:
            return Table([spec[self.key] for spec in self.specs],
                         value_dimensions=[self.key])

        table = Table(key_dimensions=dimension_labels,
                      value_dimensions=[self.key])
        keys = []
        for spec in self.specs:
            value = spec[self.key]
            key = tuple([spec[k] for k in dimension_labels])
            if key in keys:
                key_fmt = ', '.join('%s=%r' % (k,v) for (k,v) in zip(dimension_labels, key))
                self.warning('Key clash got %s (overriding)' % key_fmt)
            table[key] = value
            keys.append(key)
        return table


    def load(self, val, **kwargs):
        """
        Load the file contents into the supplied pandas dataframe or
        HoloViews Table. This allows a selection to be made over the
        metadata before loading the file contents (may be slow).
        """
        if Table and isinstance(val, Table):
            return self.load_table(val, **kwargs)
        elif DataFrame and isinstance(val, DataFrame):
            return self.load_dframe(val, **kwargs)
        else:
            raise Exception("Type %s not a DataFrame or Table." % type(val))


    def load_table(self, table):
        """
        Load the file contents into the supplied Table using the
        specified key and filetype. The input table should have the
        filenames as values which will be replaced by the loaded
        data. If data_key is specified, this key will be used to index
        the loaded data to retrive the specified item.
        """
        items,  data_keys = [], None
        for key, filename in table.items():
            data_dict = self.filetype.data(filename[0])
            current_keys = tuple(sorted(data_dict.keys()))
            values = [data_dict[k] for k in current_keys]
            if data_keys is None:
                data_keys = current_keys
            elif data_keys != current_keys:
                raise Exception("Data keys are inconsistent")
            items.append((key, values))

        return Table(items, key_dimensions=table.key_dimensions,
                     value_dimensions=data_keys)


    def load_dframe(self, dframe):
        """
        Load the file contents into the supplied dataframe using the
        specified key and filetype.
        """
        filename_series = dframe[self.key]
        loaded_data = filename_series.map(self.filetype.data)
        keys = [list(el.keys()) for el in loaded_data.values]
        for key in set().union(*keys):
            key_exists = key in dframe.columns
            if key_exists:
                self.warning("Appending '_data' suffix to data key %r to avoid"
                             "overwriting existing metadata with the same name." % key)
            suffix = '_data' if key_exists else ''
            dframe[key+suffix] = loaded_data.map(lambda x: x.get(key, np.nan))
        return dframe


    def _info(self, source, key, filetype, ignore):
        """
        Generates the union of the source.specs and the metadata
        dictionary loaded by the filetype object.
        """
        specs, mdata = [], {}
        mdata_clashes  = set()

        for spec in source.specs:
            if key not in spec:
                raise Exception("Key %r not available in 'source'." % key)

            mdata = dict((k,v) for (k,v) in filetype.metadata(spec[key]).items()
                         if k not in ignore)
            mdata_spec = dict(spec, **mdata)
            specs.append(mdata_spec)
            mdata_clashes = mdata_clashes | (set(spec.keys()) & set(mdata.keys()))
        # Metadata clashes can be avoided by using the ignore list.
        if mdata_clashes:
            self.warning("Loaded metadata keys overriding source keys.")
        return specs
